package physiology;

import java.io.Console;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.Scanner;

import nu.xom.Attribute;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.Serializer;

import dbpatient.OraclePatient;
import dbpatient.TSEvent;

public class KalmanPatient implements Serializable
{
  private static final long serialVersionUID = 1L;

  // Patient info.
  int id_;
  String insulin_type_;
  double bm_; // body mass, in kilograms.
  
  BGLValues bgl_values_;
  BasalValues basal_values_;
  BolusValues bolus_values_;
  MealValues meal_values_;
  InsulinSensitivityValues is_values_;
  CarbRatioValues cr_values_;
  
  public KalmanPatient(OraclePatient op)
  {
    id_ = op.id_;
    insulin_type_ = op.insulin_type_;
    bm_ = 0.4536 * op.weight_;
    
    bgl_values_ = new BGLValues(op.getGlucoseLevels());
    basal_values_ = new BasalValues(op.getBasalValues(), op.getTempBasalValues());
    bolus_values_ = new BolusValues(op.getBoluses());
    meal_values_ = new MealValues(op.getMeals());
    is_values_ = new InsulinSensitivityValues(op.getInsulinSensitivity());
    cr_values_ = new CarbRatioValues(op.getCarbRatio());
  }
  
  
  public void rewind()
  {
    bgl_values_.rewind();
    basal_values_.rewind();
    bolus_values_.rewind();
    meal_values_.rewind();
    is_values_.rewind();
    cr_values_.rewind();
  }
  
  
  public DeltaData nextDeltaData()
  {
    DeltaData dd = bgl_values_.nextDeltaData();
    if (dd == null)
      return null;
    
    // System.out.println("Time = " + (new TSEvent(dd.begin_)).getTS() + ", pos = " + bgl_values_.pos_);
    
    long ms = dd.end_.getTime() - dd.begin_.getTime();
    
    dd.basal_ = basal_values_.getBasal(dd.begin_, dd.end_);
    if (dd.basal_ == null) {
      System.out.println("Missing basal, using 0.0:");
      System.out.println("  Begin: " + (new TSEvent(dd.begin_)).getTS());
      System.out.println("  End: " + (new TSEvent(dd.end_)).getTS());
      dd.basal_ = 0.0;
    }
    dd.basal_ = dd.basal_ * ms / (double) (60 * 60 * 1000); // Compute quantity from 1 hour to interval.
    dd.bolus_ = bolus_values_.getBolus(dd.begin_, dd.end_);
    dd.carbs_ = meal_values_.getCarbs(dd.begin_, dd.end_);
    
    dd.insulin_sensitivity_ = is_values_.getInsulinSensitivity(dd.begin_, dd.end_);
    if (dd.insulin_sensitivity_ != null)
      dd.insulin_sensitivity_ = dd.insulin_sensitivity_ / 100;
    else
      dd.insulin_sensitivity_ = 0.5;

    dd.carb_ratio_ = cr_values_.getCarbRatio(dd.begin_, dd.end_);
    
    return dd;
  }
  
  
  public void saveIntoXML(String xml_file) throws IOException
  {
    Serializer ser = new Serializer(new FileOutputStream(xml_file, false), "UTF-8");
    ser.setIndent(2);
    ser.setMaxLength(0);

    Element e_doc = new Element("patient");
    e_doc.addAttribute(new Attribute("id", String.valueOf(id_)));
    e_doc.addAttribute(new Attribute("bm", String.valueOf(bm_)));
    e_doc.addAttribute(new Attribute("insulin_type", insulin_type_));
    
    DeltaData dd = nextDeltaData();
    while (dd != null) {
      e_doc.appendChild(getDeltaDataElement(dd));
      dd = nextDeltaData();
    }
    
    Document doc = new Document(e_doc);
    ser.write(doc);
    ser.flush();
  }
  
  
  private Element getDeltaDataElement(DeltaData dd)
  {
    Element e_dd = new Element("delta");
    e_dd.addAttribute(new Attribute("begin", (new TSEvent(dd.begin_)).getTS()));
    e_dd.addAttribute(new Attribute("end", (new TSEvent(dd.end_)).getTS()));
    
    if (dd.bgl_ != null) {
      Element e_bgl = new Element("bgl");
      e_bgl.appendChild(dd.bgl_.toString());
      e_dd.appendChild(e_bgl);
    }
    
    if (dd.basal_ != null) {
      Element e_basal = new Element("basal");
      e_basal.appendChild(dd.basal_.toString());
      e_dd.appendChild(e_basal);
    }
    
    if (dd.bolus_ != null) {
      Element e_bolus = new Element("bolus");
      e_bolus.appendChild(dd.bolus_.toString());
      e_dd.appendChild(e_bolus);      
    }
    
    if (dd.carbs_ != null) {
      Element e_carbs = new Element("meal");
      e_carbs.appendChild(dd.carbs_.toString());
      e_dd.appendChild(e_carbs);      
    }
    
    if (dd.insulin_sensitivity_ != null) {
      Element e_is = new Element("is");
      e_is.appendChild(dd.insulin_sensitivity_.toString());
      e_dd.appendChild(e_is);
    }

    if (dd.carb_ratio_ != null) {
      Element e_cr = new Element("cr");
      e_cr.appendChild(dd.carb_ratio_.toString());
      e_dd.appendChild(e_cr);      
    }
    
    return e_dd;
  }
  
  
  public static void main_xml(String args[])
  {
    if (args.length != 2) {
      System.err.println("Usage: java KalmanPatient <pid> <xml-file>!");
      System.exit(1);
    }
    
    OraclePatient op = new OraclePatient(Integer.parseInt(args[0]));
    try {
      // Read patient data from Oracle DB.
      Connection conn = op.getConnection();
      op.loadFromDB(conn);

      // Transform patient data into Kalman sequence.
      KalmanPatient kp = new KalmanPatient(op);
      kp.saveIntoXML(args[1]);

      conn.close();
    }
    catch (SQLException e) {
      e.printStackTrace();
    }
    catch (IOException e) {
      e.printStackTrace();
    }
  }

  public static void main_ser(String args[])
  {
    if (args.length != 2) {
      System.err.println("Usage: java KalmanPatient <pid> <ser-file>!");
      System.exit(1);
    }
    
    OraclePatient op = new OraclePatient(Integer.parseInt(args[0]));
    try {
      // Read patient data from Oracle DB.
      Connection conn = op.getConnection();
      op.loadFromDB(conn);

      // Transform patient data into Kalman sequence.
      KalmanPatient kp = new KalmanPatient(op);
      ObjectOutputStream out = new ObjectOutputStream(new FileOutputStream(args[1]));
      out.writeObject(kp);
      out.close();
      
      conn.close();
    }
    catch (SQLException e) {
      e.printStackTrace();
    }
    catch (IOException e) {
      e.printStackTrace();
    }
  }

  
  public static void main(String args[])
  {
    main_ser(args);
  }
}
