package physiology;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.PrintWriter;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Date;
import java.util.Vector;

import Jama.Matrix;
import kalman.ExtendedKalmanFilter;

import nu.xom.Attribute;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.Serializer;


// Kalman physiological model.
public class PhysiologicalModel extends ExtendedKalmanFilter
{
  static public double alpha_C_I_ = 0.033;
  static public double alpha_f_I_ = 0.015 ;
  static public double alpha1_clr_ = 0.043;
  static public double tau_g_renal_ = 115;
  static public double alpha1_ind_ = 1.2;
  static public double alpha1_dep_ = 0.01; // 0.0002;
  static public double alpha2_dep_ = 120; // 90; 
  static public double alpha1_C_ = 0.95;
  static public double alpha2_C_ = 1.0; // 0.09167;
  static public double alpha3_C_ = 1000; // 250;
  static public double alpha1_egp_ = -0.0333;
  static public double alpha2_egp_ = 90; // 45;
  static public double alpha3_egp_ = 40; // 15;
  static public double alpha4_egp_ = 0;
  
  static public int GLUCOSE_IDX_ = 0;
  static public int GLUCOSE_M_IDX_ = 1;
  static public int CARBS_GUT1_IDX_ = 2;
  static public int CARBS_GUT2_IDX_ = 3;
  static public int INSULIN_IDX_ = 4;
  static public int INSULIN_M_IDX_ = 5;
  static public int INSULIN_SUB_IDX_ = 6;
  static double[][] X0_;
  static double[][] P0_;
  static Matrix Q_;
  static Matrix R_;
  
  static {
    // Set initial system state.
    X0_ = new double[7][1];
    X0_[GLUCOSE_IDX_][0] = 135;
    X0_[CARBS_GUT1_IDX_][0] = 0;
    X0_[CARBS_GUT2_IDX_][0] = 0;
    X0_[INSULIN_IDX_][0] = 0;
    X0_[INSULIN_SUB_IDX_][0] = 0;
    
    // Set initial system state covariance.
    P0_ = new double[7][7];
    for (int i = 0; i < P0_.length; i++)
      Arrays.fill(P0_[i], 0.0);
    P0_[GLUCOSE_IDX_][GLUCOSE_IDX_] = 65 * 65; 
    P0_[CARBS_GUT1_IDX_][CARBS_GUT1_IDX_] = 10 * 10;
    P0_[CARBS_GUT2_IDX_][CARBS_GUT2_IDX_] = 10 * 10;
    P0_[INSULIN_IDX_][INSULIN_IDX_] = 10 * 10; 
    P0_[INSULIN_SUB_IDX_][INSULIN_SUB_IDX_] = 0.0;
    
    // Set state transition noise covariance.
    double[][] q = new double[7][7];
    for (int i = 0; i < q.length; i++)
      Arrays.fill(q[i], 0.0);
    q[GLUCOSE_IDX_][GLUCOSE_IDX_] = 65 * 65; 
    q[GLUCOSE_M_IDX_][GLUCOSE_M_IDX_] = 65 * 65 * 150 * 150; 
    q[CARBS_GUT1_IDX_][CARBS_GUT1_IDX_] = 10 * 10;
    q[CARBS_GUT2_IDX_][CARBS_GUT2_IDX_] = 10 * 10;
    q[INSULIN_IDX_][INSULIN_IDX_] = 10 * 10;
    q[INSULIN_M_IDX_][INSULIN_M_IDX_] = 10 * 10 * 10000 * 10000;
    q[INSULIN_SUB_IDX_][INSULIN_SUB_IDX_] = 100 * 100;
    /*
    q[GLUCOSE_IDX_][GLUCOSE_IDX_] = 0.1 * 0.1;
    q[GLUCOSE_IDX_][GLUCOSE_M_IDX_] = q[GLUCOSE_M_IDX_][GLUCOSE_IDX_] = 0.1 * 10;
    q[GLUCOSE_IDX_][CARBS_GUT2_IDX_] = q[CARBS_GUT2_IDX_][GLUCOSE_IDX_] = 0.1 * 0.05;
    q[GLUCOSE_IDX_][INSULIN_IDX_] = q[INSULIN_IDX_][GLUCOSE_IDX_] = 0.1 * 10;
    q[GLUCOSE_M_IDX_][GLUCOSE_M_IDX_] = 0.1 * 0.1 * 10 * 10;
    q[GLUCOSE_M_IDX_][CARBS_GUT2_IDX_] = q[CARBS_GUT2_IDX_][GLUCOSE_M_IDX_] = 0.1 * 0.05 * 10;
    q[GLUCOSE_M_IDX_][INSULIN_IDX_] = q[INSULIN_IDX_][GLUCOSE_M_IDX_] = 10 * 10;
    q[CARBS_GUT1_IDX_][CARBS_GUT1_IDX_] = 0.05 * 0.05;
    q[CARBS_GUT2_IDX_][CARBS_GUT1_IDX_] = q[CARBS_GUT1_IDX_][CARBS_GUT2_IDX_] = 0.05 * 0.05;
    q[CARBS_GUT2_IDX_][CARBS_GUT2_IDX_] = 0.05 * 0.05;
    q[INSULIN_IDX_][INSULIN_IDX_] = 10 * 10;
    q[INSULIN_IDX_][INSULIN_SUB_IDX_] = q[INSULIN_SUB_IDX_][INSULIN_IDX_] = 10 * 100;
    q[INSULIN_M_IDX_][INSULIN_M_IDX_] = 100 * 100;
    q[INSULIN_IDX_][INSULIN_M_IDX_] = q[INSULIN_M_IDX_][INSULIN_IDX_] = 10 * 100;
    q[INSULIN_M_IDX_][INSULIN_SUB_IDX_] = q[INSULIN_SUB_IDX_][INSULIN_M_IDX_] = 100 * 100;
    q[INSULIN_SUB_IDX_][INSULIN_SUB_IDX_] = 100 * 100;
    */
    Q_ = new Matrix(q);
    
    // Set observation noise covariance.
    double[][] r = new double[2][2];
    r[0][0] = 0.1 * 0.1;
    r[1][1] = 10 * 10;
    r[0][1] = r[1][0] = 0.0;
    R_ = new Matrix(r);
  }
  
  KalmanPatient kp_;
  DeltaData dd_;
  
  public PhysiologicalModel(KalmanPatient kp)
  {
    kp_ = kp;
    // TO BE REMOVED !!!
    kp_.bm_ = 80;
    
    X0_[GLUCOSE_M_IDX_][0] = X0_[GLUCOSE_IDX_][0] * 2.2 * kp_.bm_;
    X0_[INSULIN_M_IDX_][0] = X0_[INSULIN_IDX_][0] * 142 * kp_.bm_;

    P0_[GLUCOSE_M_IDX_][GLUCOSE_M_IDX_] = 
        P0_[GLUCOSE_IDX_][GLUCOSE_IDX_] * (2.2 * kp_.bm_) * (2.2 * kp_.bm_);
    P0_[INSULIN_M_IDX_][INSULIN_M_IDX_] = 
        P0_[INSULIN_IDX_][INSULIN_IDX_] * (142 * kp_.bm_) * (142 * kp_.bm_);
    
    init();
  }
  
  
  public void init()
  {
    kp_.rewind();
    
    t_ = 0;
    X_ = Matrix.constructWithCopy(X0_);
    P_ = Matrix.constructWithCopy(P0_);
  }
  
  
  public Matrix computeNextState()
  {
    double[][] X = new double[7][1];
    X[GLUCOSE_IDX_][0] = computeNextGlucose();
    X[GLUCOSE_M_IDX_][0] = computeNextGlucoseMass();
    X[CARBS_GUT1_IDX_][0] = computeNextCarbsGut1();
    X[CARBS_GUT2_IDX_][0] = computeNextCarbsGut2();
    X[INSULIN_IDX_][0] = computeNextInsulin();
    X[INSULIN_M_IDX_][0] = computeNextInsulinMass();
    X[INSULIN_SUB_IDX_][0] = computeNextInsulinSub();
    
    return new Matrix(X);
  }

  
  double computeNextGlucoseMass()
  {
    double g = X_.get(GLUCOSE_IDX_, 0);
    double gm = X_.get(GLUCOSE_M_IDX_, 0);
    double I = X_.get(INSULIN_IDX_, 0);
    double C_gut2 = X_.get(CARBS_GUT2_IDX_, 0);
    
    // Endogenous production \delta_g_t_egp.
    double delta_egp = Math.max(0, alpha1_egp_ * g + alpha2_egp_ * Math.exp(- I / alpha3_egp_) - alpha4_egp_);
    
    // Meal absorption \delta_g_t_gut.
    double delta_gut = alpha3_C_ * alpha2_C_ * C_gut2 / (C_gut2 + 25);
    
    // Insulin independent utilization \delta_g_t_ind.
    double delta_ind = alpha1_ind_ * Math.sqrt(g);
    
    // Insulin dependent utilization \delta_g_t_dep.
    double delta_dep = alpha1_dep_ * I * (g + alpha2_dep_);
    
    double delta_clr = 0.0;
    // Renal clearance \delta_g_t_clr.
    if (g > tau_g_renal_)
      delta_clr = alpha1_clr_ * (g - tau_g_renal_);
    
    return gm + delta_egp + delta_gut - delta_ind - delta_dep - delta_clr;
  }
  
  
  double computeNextGlucose()
  {
    double gm = computeNextGlucoseMass();
        
    return gm / (2.2 * kp_.bm_);
  }
  
  
  double computeNextCarbsGut1()
  {
    double Cg1 = X_.get(CARBS_GUT1_IDX_, 0);
    
    double cgut1 = (1 - alpha1_C_) * Cg1;
    if (dd_.carbs_ != null)
      cgut1 += dd_.carbs_;
    
    return cgut1;
  }
  
  
  double computeNextCarbsGut2()
  {
    double Cg1 = X_.get(CARBS_GUT1_IDX_, 0);
    double Cg2 = X_.get(CARBS_GUT2_IDX_, 0);
    
    double cgut2 = (1 - alpha2_C_ / (Cg2 + 25)) * Cg2 + alpha1_C_ * Cg1;
    
    return cgut2;
  }
  
  
  double computeNextInsulinMass()
  {
    double Im = X_.get(INSULIN_M_IDX_, 0);
    double Isub = X_.get(INSULIN_SUB_IDX_, 0);

    double inew = Im * (1 - alpha_C_I_) + Isub * alpha_f_I_;
    
    return inew;
  }
  
  
  double computeNextInsulin()
  {
    double Im = computeNextInsulinMass();
    dd_.insulin_sensitivity_ = 1.0; // TO BE REMOVED!
    return Im * dd_.insulin_sensitivity_ * 1.0 / (142 * kp_.bm_);
  }
  
  
  double computeNextInsulinSub()
  {
    double Isub = X_.get(INSULIN_SUB_IDX_, 0);

    double isubnew = Isub * (1 - alpha_f_I_);
    if (dd_.basal_ != null)
      isubnew += dd_.basal_ * 1000 * 1000; // transform in microunits.
    if (dd_.bolus_ != null)
      isubnew += dd_.bolus_ * 1000 * 1000; // transform in microunits.
    
    return isubnew;
  }
  
  
  public Matrix getStateCovarianceQ()
  {
    return Q_;
  }

  
  public Matrix getStateJacobianF()
  {
    double[][] J = new double[7][7];
    for (int i = 0; i < J.length; i++)
      Arrays.fill(J[i], 0);
    
    double g = X_.get(GLUCOSE_IDX_, 0);
    double I = X_.get(INSULIN_IDX_, 0);
    double Cg2 = X_.get(CARBS_GUT2_IDX_, 0);

    // Glucose mass.
    J[GLUCOSE_M_IDX_][GLUCOSE_IDX_] =
        alpha1_egp_ - alpha1_ind_ / (2 * Math.sqrt(g)) - alpha1_dep_ * I;
    if (g > tau_g_renal_)
      J[GLUCOSE_M_IDX_][GLUCOSE_IDX_] -= alpha1_clr_;
    J[GLUCOSE_M_IDX_][GLUCOSE_M_IDX_] = 1.0;
    J[GLUCOSE_M_IDX_][CARBS_GUT2_IDX_] = 
        alpha3_C_ * alpha2_C_ * 25 / ((Cg2 + 25) * (Cg2 + 25));
    J[GLUCOSE_M_IDX_][INSULIN_IDX_] = 
        - alpha2_egp_ / alpha3_egp_ * Math.exp(- I / alpha3_egp_) 
        - alpha1_dep_ * (g + alpha2_dep_);
    
    // Glucose concentration.
    for (int i = 0; i < J.length; i++)
      J[GLUCOSE_IDX_][i] = J[GLUCOSE_M_IDX_][i] / (2.2 * kp_.bm_);
    
    // Carbs gut compartment 1.
    J[CARBS_GUT1_IDX_][CARBS_GUT1_IDX_] = 1 - alpha1_C_;
    
    // Carbs gut compartment 2.
    J[CARBS_GUT2_IDX_][CARBS_GUT1_IDX_] = alpha1_C_;
    J[CARBS_GUT2_IDX_][CARBS_GUT2_IDX_] = (1 - alpha2_C_) * 25 / ((Cg2 + 25) * (Cg2 + 25));
    
    // Insulin mass.
    J[INSULIN_M_IDX_][INSULIN_M_IDX_] = 1 - alpha_C_I_;    
    J[INSULIN_M_IDX_][INSULIN_SUB_IDX_] = alpha_f_I_;
    
    // Insulin concentration.
    dd_.insulin_sensitivity_ = 1.0; // TO BE REMOVED!
    for (int i = 0; i < J.length; i++)
      J[INSULIN_IDX_][i] = J[INSULIN_M_IDX_][i] * dd_.insulin_sensitivity_ * 1.0 / (142 * kp_.bm_);

    // Insulin subcutaneous.
    J[INSULIN_SUB_IDX_][INSULIN_SUB_IDX_] = 1 - alpha_f_I_;
    
    return new Matrix(J);
  }

  
  public Matrix computeObservation()
  {
    double[][] bgl = new double[2][1];
    
    bgl[0][0] = X_.get(GLUCOSE_IDX_, 0);
    bgl[1][0] = X_.get(GLUCOSE_M_IDX_, 0);
    
    return new Matrix(bgl);
  }

  
  public Matrix getObsCovarianceR()
  {
    return R_;
  }

  public Matrix getObsJacobianH()
  {
    double[][] J = new double[2][7];
    Arrays.fill(J[0], 0.0);
    Arrays.fill(J[1], 0.0);
    J[0][GLUCOSE_IDX_] = 1;
    J[1][GLUCOSE_M_IDX_] = 1;
    
    return new Matrix(J);
  }
  
  
  public PhysiologicalHistory filter()
  {
    init();
    
    PhysiologicalHistory ph = new PhysiologicalHistory();
    
    int steps = 0;
    double last_bgl = 100.0;
    
    dd_ = kp_.nextDeltaData();
    while (dd_ != null) {
      // Use physiological model equations to update state X_ and P_.
      System.out.println(dd_);
      System.out.println(">>> Prediction:");
      predictState();
      //forceConstraints();
      X_.print(6, 2);
      P_.print(6, 2);
      // If observation available, use it to correct state X_ and P_.
      if (dd_.bgl_ != null) {
        double[][] bgl = new double[2][1];
        bgl[0][0] = dd_.bgl_;
        bgl[1][0] = dd_.bgl_ * (2.2 * kp_.bm_);
        correctState(new Matrix(bgl));
        forceConstraints();
        System.out.println(">>> Correction:");
        X_.print(6, 2);
        P_.print(6, 2);
        
        steps = 0;
        last_bgl = dd_.bgl_;
      }
      else {
        // If after 10 steps there is no observation, use last observation.
        steps++;
        if (steps >= 10) {
          double[][] bgl = new double[2][1];
          bgl[0][0] = last_bgl;
          bgl[1][0] = last_bgl * (2.2 * kp_.bm_);
          correctState(new Matrix(bgl));
          forceConstraints();
          System.out.println(">>> Correction to last observed BGL:");
          X_.print(6, 2);
          P_.print(6, 2);
          
          steps = 0;
        }
      }
      // Memorize system transition path.
      ph.addStep(dd_, X_, P_);
      // Next time step.
      dd_ = kp_.nextDeltaData();
    }
    
    return ph;
  }
  
  
  protected void forceConstraints()
  {
    double[][] X = X_.getArray();
    for (int i = 0; i < 7; i++) {
      if (X[i][0] < 0)
        X[i][0] = 0;
    }
  }
  
  
  public Vector<PhysiologicalHistory> predict(PhysiologicalHistory ph, Vector<Date> times, int steps)
  {
    Vector<PhysiologicalHistory> predictions = new Vector<PhysiologicalHistory>();
    int pos = 0;
    // Assume times and ph are ordered on timestamps.
    for (Date ts: times) {
      System.out.println(";;; Prediction for timestamp " + ts);
      // Find timestamp position in physiological history ph.
      while (pos < ph.steps_.size() && ph.steps_.get(pos).dd_.end_.compareTo(ts) < 0)
        pos++;
      assert ph.steps_.get(pos).dd_.end_.equals(ts);
      PhysiologicalHistory php = predict(ph, pos, steps);
      predictions.add(php);
      
      pos++;
    }
    
   return predictions;
  }
  
  
  public PhysiologicalHistory predict(PhysiologicalHistory ph, int pos, int steps)
  {
    PhysiologicalHistory php = new PhysiologicalHistory();
    
    PhysiologicalHistory.PhysiologicalStep step = ph.steps_.get(pos);
    X_ = step.X_;
    P_ = step.P_;
    dd_ = step.dd_;
    // Memorize start state.
    php.addStep(dd_, X_, P_);
        
    for (int i = 1; i <= steps; i++) {
      // Next time step.
      step = ph.steps_.get(pos + i);
      dd_ = step.dd_;
      // Use physiological model equations to update state X_ and P_.
      System.out.println(dd_);
      System.out.println(">>> Prediction:");
      predictState();
      X_.print(6, 2);
      P_.print(6, 2);
      // Memorize system transition path.
      php.addStep(dd_, X_, P_);
    }
    
    return php;
  }
  
  
  static public Vector<Date> loadTimeStamps(String file)
  {
    SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
    
    Vector<Date> times = new Vector<Date>();
    try {
      BufferedReader br = new BufferedReader(new FileReader(file));
      String line = br.readLine();
      while (line != null) {
        times.add(sdf.parse(line));
        line = br.readLine();
      }
    }
    catch (IOException e) {
      e.printStackTrace();
      System.exit(1);
    }
    catch (ParseException e) {
      e.printStackTrace();
      System.exit(2);
    }
    
    return times;
  }
  
  
  static public double computeRMSE(Vector<PhysiologicalHistory> predictions, int steps)
  {
    double rmse = 0;
    int count = 0;
    for (PhysiologicalHistory php: predictions) {
      PhysiologicalHistory.PhysiologicalStep step = php.steps_.get(steps);
      if (step.dd_.bgl_ != null) {
        double bgl_system = step.X_.get(GLUCOSE_IDX_, 0);
        double bgl_true = step.dd_.bgl_;
        rmse += (bgl_true - bgl_system) * (bgl_true - bgl_system);
        count++;
      }
      else
        System.out.println("[!] Missing BGL measurement at timestamp " + step.dd_.end_);
    }
    rmse = Math.sqrt(rmse / count);
    
    return rmse;
  }
  
  
  static public void main_filter(String[] args)
  {
    if (args.length != 3) {
      System.err.println("Usage: java PhysiologicalModel filter <in:ser-file> <out:xml-file>!");
      System.exit(1);
    }

    try {
      ObjectInputStream is = new ObjectInputStream(new FileInputStream(args[1]));
      KalmanPatient kp = (KalmanPatient) is.readObject();
      is.close();
      PhysiologicalModel pm = new PhysiologicalModel(kp);
      PhysiologicalHistory ph = pm.filter();
      pm.savePhysiologicalHistory(ph, args[2]);
    }
    catch (IOException e) {
      e.printStackTrace();
      System.exit(2);
    }
    catch (ClassNotFoundException e) {
      e.printStackTrace();
      System.exit(3);
    }
  }
  
  
  public void savePhysiologicalHistory(PhysiologicalHistory ph, String xml_file)
    throws IOException
  {
    Serializer ser = new Serializer(new FileOutputStream(xml_file, false), "UTF-8");
    ser.setIndent(2);
    ser.setMaxLength(0);

    Element e_doc = new Element("patient");
    e_doc.addAttribute(new Attribute("id", String.valueOf(kp_.id_)));

    SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");

    for (PhysiologicalHistory.PhysiologicalStep step: ph.steps_) {
      Element e_step = new Element("step");
      e_step.addAttribute(new Attribute("ts", sdf.format(step.dd_.end_)));
      
      Element e_carbs = new Element("carbs");
      e_carbs.addAttribute(new Attribute("gut1", 
                                         String.valueOf(step.X_.get(CARBS_GUT1_IDX_, 0))));
      e_carbs.addAttribute(new Attribute("gut2",
                                         String.valueOf(step.X_.get(CARBS_GUT2_IDX_, 0))));
      e_step.appendChild(e_carbs); 
      
      Element e_insulin = new Element("insulin");
      e_insulin.addAttribute(new Attribute("concentration",
                                           String.valueOf(step.X_.get(INSULIN_IDX_, 0))));
      e_insulin.addAttribute(new Attribute("mass",
                                           String.valueOf(step.X_.get(INSULIN_M_IDX_, 0))));
      e_insulin.addAttribute(new Attribute("subcutaneous",
                                           String.valueOf(step.X_.get(INSULIN_SUB_IDX_, 0))));
      e_step.appendChild(e_insulin);
      
      e_doc.appendChild(e_step);
    }
     
    Document doc = new Document(e_doc);
    ser.write(doc);
    ser.flush();

  }


  static public void savePredictions(Vector<PhysiologicalHistory> predictions, String file)
  {
    SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
    try {
      PrintWriter pw = new PrintWriter(file);
      for (PhysiologicalHistory php: predictions) {
        PhysiologicalHistory.PhysiologicalStep step0 = php.steps_.get(0);
        pw.println(sdf.format(step0.dd_.end_));
        pw.println("  State at 0 minutes:");
        for (int s = 0; s < 7; s++)
          pw.println("    " + step0.X_.get(s, 0));
        
	PhysiologicalHistory.PhysiologicalStep step05 = php.steps_.get(5);
        pw.println("  State at 05 minutes:");
        for (int s = 0; s < 7; s++)
          pw.println("    " + step05.X_.get(s, 0));

	PhysiologicalHistory.PhysiologicalStep step10 = php.steps_.get(10);
        pw.println("  State at 10 minutes:");
        for (int s = 0; s < 7; s++)
          pw.println("    " + step10.X_.get(s, 0));

	PhysiologicalHistory.PhysiologicalStep step15 = php.steps_.get(15);
        pw.println("  State at 15 minutes:");
        for (int s = 0; s < 7; s++)
          pw.println("    " + step15.X_.get(s, 0));

	PhysiologicalHistory.PhysiologicalStep step20 = php.steps_.get(20);
        pw.println("  State at 20 minutes:");
        for (int s = 0; s < 7; s++)
          pw.println("    " + step20.X_.get(s, 0));

	PhysiologicalHistory.PhysiologicalStep step25 = php.steps_.get(25);
        pw.println("  State at 25 minutes:");
        for (int s = 0; s < 7; s++)
          pw.println("    " + step25.X_.get(s, 0));

        PhysiologicalHistory.PhysiologicalStep step30 = php.steps_.get(30);
        pw.println("  State at 30 minutes:");
        for (int s = 0; s < 7; s++)
          pw.println("    " + step30.X_.get(s, 0));
        
	PhysiologicalHistory.PhysiologicalStep step35 = php.steps_.get(35);
        pw.println("  State at 35 minutes:");
        for (int s = 0; s < 7; s++)
          pw.println("    " + step35.X_.get(s, 0));

	PhysiologicalHistory.PhysiologicalStep step40 = php.steps_.get(40);
        pw.println("  State at 40 minutes:");
        for (int s = 0; s < 7; s++)
          pw.println("    " + step40.X_.get(s, 0));

	PhysiologicalHistory.PhysiologicalStep step45 = php.steps_.get(45);
        pw.println("  State at 45 minutes:");
        for (int s = 0; s < 7; s++)
          pw.println("    " + step45.X_.get(s, 0));

	PhysiologicalHistory.PhysiologicalStep step50 = php.steps_.get(50);
        pw.println("  State at 50 minutes:");
        for (int s = 0; s < 7; s++)
          pw.println("    " + step50.X_.get(s, 0));

	PhysiologicalHistory.PhysiologicalStep step55 = php.steps_.get(55);
        pw.println("  State at 55 minutes:");
        for (int s = 0; s < 7; s++)
          pw.println("    " + step55.X_.get(s, 0));	
	
        PhysiologicalHistory.PhysiologicalStep step60 = php.steps_.get(60);
        pw.println("  State at 60 minutes:");
        for (int s = 0; s < 7; s++)
          pw.println("    " + step60.X_.get(s, 0));
      }
      pw.close();
    }
    catch (IOException e) {
      e.printStackTrace();
      System.exit(1);
    }
  }
  

  static public void savePredictedBGLs(Vector<PhysiologicalHistory> predictions, String file)
  {
    SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
    try {
      PrintWriter pw = new PrintWriter(file);
      for (PhysiologicalHistory php: predictions) {
        PhysiologicalHistory.PhysiologicalStep step0 = php.steps_.get(0);
        pw.println(sdf.format(step0.dd_.end_));
        for (int i = 0; i < 12; i++) {
          PhysiologicalHistory.PhysiologicalStep step = php.steps_.get(i * 5);
          pw.println("  " + step.X_.get(GLUCOSE_IDX_, 0));
        }
      }
      pw.close();
    }
    catch (IOException e) {
      e.printStackTrace();
      System.exit(1);
    }
  }
  

  static public void main_predict(String[] args)
  {
    if (args.length != 4) {
      System.err.println("Usage: java PhysiologicalModel predict <in:ser-file> <in:ts-file> <out:pred-file>!");
      System.exit(1);
    }

    try {
      ObjectInputStream is = new ObjectInputStream(new FileInputStream(args[1]));
      KalmanPatient kp = (KalmanPatient) is.readObject();
      is.close();
      
      PhysiologicalModel pm = new PhysiologicalModel(kp);
      System.out.println("Filtering ...");
      PhysiologicalHistory ph = pm.filter();
      
      Vector<Date> times = PhysiologicalModel.loadTimeStamps(args[2]);
      System.out.println("Prediction ...");
      Vector<PhysiologicalHistory> predictions = pm.predict(ph, times, 12 * 5);
      
      double rmse_30 = computeRMSE(predictions, 6 * 5);
      double rmse_60 = computeRMSE(predictions, 12 * 5);
      System.out.println("RMSE(30) = " + rmse_30);
      System.out.println("RMSE(60) = " + rmse_60);
      
      savePredictions(predictions, args[3]);
      //savePredictedBGLs(predictions, args[3]);
    }
    catch (IOException e) {
      e.printStackTrace();
      System.exit(2);
    }
    catch (ClassNotFoundException e) {
      e.printStackTrace();
      System.exit(3);
    }
  }



  static public void main(String[] args)
  {
    if (args.length == 0) {
      System.err.println("Usage: java PhysiologicalModel <filter|predict> ...");
      System.exit(1);
    }

    if (args[0].equalsIgnoreCase("filter"))
      main_filter(args);
    else if (args[0].equalsIgnoreCase("predict"))
      main_predict(args);
    else {
      System.err.println("Usage: java PhysiologicalModel <filter|predict> ...");
      System.exit(1);
    }
  }

}
