package physiology;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.util.Date;
import java.util.Vector;

import Jama.Matrix;

// Deterministic physiological model.
public class PhysiologicalModelD extends PhysiologicalModel
{
  public PhysiologicalModelD(KalmanPatient kp)
  {
    super(kp);
  }
  
  
  public void predictState()
  {
    // Update mean X of state vector.
    X_ = computeNextState();

    // Update time.
    t_++;
  }

  
  public void correctState(Matrix O)
  {
    X_.set(GLUCOSE_IDX_, 0, O.get(0, 0));
    X_.set(GLUCOSE_M_IDX_, 0, O.get(0, 0) * 2.2 * kp_.bm_);
  }


  public PhysiologicalHistory filter()
  {
    init();
    
    PhysiologicalHistory ph = new PhysiologicalHistory();
    
    dd_ = kp_.nextDeltaData();
    while (dd_ != null) {
      // Use physiological model equations to update state X_.
      System.out.println(dd_);
      System.out.println(">>> Prediction:");
      predictState();
      //forceConstraints();
      X_.print(6, 2);
      // If observation available, use it to correct state X_.
      if (dd_.bgl_ != null) {
        double[][] bgl = new double[2][1];
        bgl[0][0] = dd_.bgl_;
        bgl[1][0] = dd_.bgl_ * (2.2 * kp_.bm_);
        correctState(new Matrix(bgl));
        forceConstraints();
        System.out.println(">>> Correction:");
        X_.print(6, 2);
      }
      // Memorize system transition path.
      ph.addStep(dd_, X_, P_);
      // Next time step.
      dd_ = kp_.nextDeltaData();
    }
    
    return ph;
  }
  
  
  static public void main_filter(String[] args)
  {
    if (args.length != 1) {
      System.err.println("Usage: java PhysiologicalModelD <ser-file>!");
      System.exit(1);
    }

    try {
      ObjectInputStream is = new ObjectInputStream(new FileInputStream(args[0]));
      KalmanPatient kp = (KalmanPatient) is.readObject();
      is.close();
      PhysiologicalModel pm = new PhysiologicalModelD(kp);
      pm.filter();
    }
    catch (IOException e) {
      e.printStackTrace();
      System.exit(2);
    }
    catch (ClassNotFoundException e) {
      e.printStackTrace();
      System.exit(3);
    }
  }
  
  
  static public void main_predict(String[] args)
  {
    if (args.length != 3) {
      System.err.println("Usage: java PhysiologicalModelD <ser-file> <ts-file> <pred-file>!");
      System.exit(1);
    }

    try {
      ObjectInputStream is = new ObjectInputStream(new FileInputStream(args[0]));
      KalmanPatient kp = (KalmanPatient) is.readObject();
      is.close();
      
      PhysiologicalModel pm = new PhysiologicalModelD(kp);
      System.out.println("Filtering ...");
      PhysiologicalHistory ph = pm.filter();
      
      Vector<Date> times = PhysiologicalModel.loadTimeStamps(args[1]);
      System.out.println("Prediction ...");
      Vector<PhysiologicalHistory> predictions = pm.predict(ph, times, 12 * 5);
      
      double rmse_30 = computeRMSE(predictions, 6 * 5);
      double rmse_60 = computeRMSE(predictions, 12 * 5);
      System.out.println("RMSE(30) = " + rmse_30);
      System.out.println("RMSE(60) = " + rmse_60);
      
      savePredictions(predictions, args[2]);
    }
    catch (IOException e) {
      e.printStackTrace();
      System.exit(2);
    }
    catch (ClassNotFoundException e) {
      e.printStackTrace();
      System.exit(3);
    }
  }
  
  static public void main(String[] args)
  {
    main_predict(args);
  }
  
}
